#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-08
# @Filename: stop.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from contextlib import suppress

from typing import TYPE_CHECKING

import click

from lvmguider.actor import lvmguider_parser
from lvmguider.maskbits import GuiderStatus


if TYPE_CHECKING:
    from lvmguider.actor import GuiderCommand


__all__ = ["stop"]


@lvmguider_parser.command()
@click.option("--now", is_flag=True, help="Aggressively stops the loop.")
async def stop(command: GuiderCommand, now=False):
    """Stops the guide loop."""

    status = command.actor.status
    if status & GuiderStatus.IDLE:
        return command.finish("Guider is not active.")

    if now:
        if command.actor.guide_task and not command.actor.guide_task.done():
            command.actor.guide_task.cancel()
            with suppress(asyncio.CancelledError):
                await command.actor.guide_task
        command.actor.guider = None
        return command.finish("Guider was forcibly stopped.")

    if command.actor.status & GuiderStatus.STOPPING:
        return command.fail("Guider loop is already stopping.")

    command.actor.status |= GuiderStatus.STOPPING
    return command.finish("Stopping the guide loop.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-19
# @Filename: guider.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import pathlib

from typing import TYPE_CHECKING, Any, cast

import numpy
import pandas
from astropy.io import fits
from astropy.table import Table
from astropy.wcs import WCS
from simple_pid import PID

from sdsstools.time import get_sjd

from lvmguider.maskbits import GuiderStatus

from .tools import get_proc_path, run_in_executor
from .transformations import (
    XZ_FULL_FRAME,
    calculate_guide_offset,
    delta_radec2mot_axis,
    solve_from_files,
    wcs_from_single_cameras,
)


if TYPE_CHECKING:
    from lvmguider.actor import GuiderCommand
    from lvmguider.astrometrynet import AstrometrySolution


class CriticalGuiderError(Exception):
    """An exception that should stop the guide loop."""

    pass


class Guider:
    """Acquisition and guiding sequence.

    Parameters
    ----------
    command
        The actor command used to communicate with the actor system.
    field_centre
        The field centre on which we want to guider.
    pixel
        The ``(x,y)`` pixel of the master frame to use to determine the pointing.
        Default to the central pixel.

    """

    def __init__(
        self,
        command: GuiderCommand,
        field_centre: tuple[float, float],
        pixel: tuple[float, float] | None = None,
    ):
        self.command = command
        self.telescope = command.actor.telescope
        self.cameras = command.actor.cameras

        self.field_centre = field_centre
        self.pixel = pixel or XZ_FULL_FRAME

        self.config = command.actor.config

        self._mode: str = "auto"

        # Use negative values because we apply the correction in the
        # same direction as the measured offset between pointing and
        # reference position.
        self.pid_ra = PID(
            Kp=-self.config["pid"]["ra"]["Kp"],
            Ki=-self.config["pid"]["ra"]["Ki"],
            Kd=-self.config["pid"]["ra"]["Kd"],
        )

        self.pid_dec = PID(
            Kp=-self.config["pid"]["dec"]["Kp"],
            Ki=-self.config["pid"]["dec"]["Ki"],
            Kd=-self.config["pid"]["dec"]["Kd"],
        )

        self.use_reference_frames: bool = False
        self.reference_frames: dict[str, pathlib.Path] = {}
        self.reference_sources: pandas.DataFrame | None = None
        self.reference_wcs: WCS | None = None

    def set_reference_frames(
        self,
        frameno: int | None = None,
        mf_wcs: WCS | None = None,
    ):
        """Sets the reference images to be used for guiding."""

        # Reset to not using reference frames.
        if frameno is None:
            self.use_reference_frames = False
            self.reference_frames = {}
            self.reference_sources = None
            self.reference_wcs = None
            return

        # Gets images that match frameno.
        sjd = get_sjd("LCO")
        agcam_dir = pathlib.Path(f"/data/agcam/{sjd}")
        files = list(agcam_dir.glob(f"lvm.{self.telescope}.agcam.*_{frameno:08d}.*"))

        if len(files) == 0:
            raise FileNotFoundError(f"No files found for frame number {frameno}.")

        if self.telescope != "spec" and len(files) == 1:
            self.command.warning(f"Only one file found for frame number {frameno}.")

        sources: list[pandas.DataFrame] = []
        for file_ in files:
            try:
                header = fits.getheader(file_)
                file_sources = Table.read(file_, "SOURCES").to_pandas()
            except Exception as err:
                raise ValueError(f"Failed retrieving sources from file {file_}: {err}")

            camname = header["CAMNAME"]
            sources.append(file_sources)
            self.reference_frames[camname] = file_

        if mf_wcs is None:
            proc_path = get_proc_path(files[0])
            if not proc_path.exists():
                raise CriticalGuiderError("Cannot retrieve reference frame master WCS.")
            self.reference_wcs = WCS(fits.getheader(str(proc_path), "ASTROMETRY"))
        else:
            self.reference_wcs = mf_wcs

        self.reference_sources = pandas.concat(sources)
        self.use_reference_frames = True

    def set_pixel(self, pixel_x: float | None = None, pixel_z: float | None = None):
        """Sets the master frame pixel coordinates ``(x, z)`` on which to guide."""

        if pixel_x is None or pixel_z is None:
            new_pixel = XZ_FULL_FRAME
        else:
            new_pixel = (pixel_x, pixel_z)

        if self.pixel != new_pixel and self._mode == "auto":
            # In auto mode, if we change the current pixel, we start drifting
            # towards a new target position.
            self.command.actor.status |= GuiderStatus.DRIFTING
            self.pixel = new_pixel
            self.set_reference_frames()

        return self.pixel

    async def guide_one(
        self,
        exposure_time: float = 5.0,
        mode: str = "auto",
        guide_tolerance: float | None = None,
        apply_correction: bool = True,
        use_motor_offsets: bool = True,
        use_individual_images: bool = True,
    ):
        """Performs one guide iteration.

        This is the main routine that executes one acquisition or guide iteration.
        When ``mode='auto'``, the code will do an acquisition step unless
        ``use_reference_frames=True``. At the end of the acquisition step, if
        the measured offset is smaller than ``guide_tolerance``, it will set
        ``use_reference_frames`` and the following iterations will be performed
        using the guide mechanism.

        During acquisition, in order, it does the following:

        - Expose the cameras. Subtract dark frame.
        - Extract sources.
        - Create a master frame based on the camera(s)-to-IFU metrology.
        - Solve the frame using astrometry.net.
        - Determine the offsets for a given pixel on the master frame. By default
          the central pixel is used, which corresponds to guiding at the central
          fibre in the IFU, but a different location can be used to guide, for example,
          on one of the spec mask holes.
        - Send PID-corrected offsets to the telescopes.

        Guiding is similar but instead of solving a new master frame WCS, the
        extracted sources are compared with the list of sources from the reference
        frame, and associated using nearest-neighbours. From there, an offset is
        derived and sent to the telescope.

        Parameters
        ----------
        exposure_time
            The exposure time, in seconds.
        mode
            Either ``'auto'``, ``'guide'``, or ``'acquire'``. In auto mode acquisition
            will be performed until the measured offset is smaller than
            ``guide_tolerance``. The other two modes force continuous acquisition
            or guiding.
        guide_tolerance
            The separation between field RA/Dec and measured pointing at which
            to consider than acquisition has been completed and guiding begins.
        apply_corrections
            Whether to apply the measured corrections. If `False`, only measures.
        use_motor_offsets
            If `True`, applies the corrections as motor axis offsets.
        use_individual_images
            If `True`, the WCS for the master frame in acquisition is derived from
            the individual WCS of the AG frames.

        """

        if self.use_reference_frames and self.reference_sources is None:
            raise CriticalGuiderError("No reference sources defined.")

        filenames, frameno, sources = await self.cameras.expose(
            self.command,
            extract_sources=True,
            exposure_time=exposure_time,
        )
        if sources is None:
            raise RuntimeError("No sources found.")

        self.command.actor.status |= GuiderStatus.PROCESSING
        self.command.actor.status &= ~GuiderStatus.IDLE

        guide_tolerance = guide_tolerance or self.config.get("guide_tolerance", 5)

        if mode == "auto":
            is_acquisition = bool(
                not self.use_reference_frames
                or self.command.actor.status & GuiderStatus.DRIFTING
            )
        elif mode == "guide":
            is_acquisition = False
        elif mode == "acquire":
            is_acquisition = True
        else:
            raise CriticalGuiderError(f"Invalid mode {mode}.")

        self._mode = mode

        if is_acquisition:
            try:
                ra_p, dec_p, wcs = await self.determine_pointing(
                    filenames,
                    pixel=self.pixel,
                    use_individual_images=use_individual_images,
                )

                # Offset is field centre - current pointing.
                pnt = (ra_p, dec_p)
                offset_radec, offset_motax, sep = self.calculate_telescope_offset(pnt)

            except Exception as err:
                raise RuntimeError(f"Failed determining telescope pointing: {err}")

        else:
            wcs = self.reference_wcs
            offset_radec, sep = await self.calculate_guide_offset(sources)

            # Fro typing. This is safe, calculate_guide_offset also checks.
            assert self.reference_wcs is not None

            # Pointing from the reference frame.
            pixel = self.pixel or XZ_FULL_FRAME
            ref_pointing = self.reference_wcs.pixel_to_world(*pixel)
            ra_ref = ref_pointing.ra.deg
            dec_ref = ref_pointing.dec.deg

            # Current pointing. Note that the offset is to go from current position
            # to reference position.
            cos_dec = numpy.cos(numpy.radians(dec_ref))
            ra_p = numpy.round(ra_ref - offset_radec[0] / 3600 * cos_dec, 6)
            dec_p = numpy.round(dec_ref - offset_radec[1] / 3600.0, 6)

            if self.command.actor.status & GuiderStatus.DRIFTING:
                raise ValueError("Guider is drifting. Reverting to acquisition.")

            if (sep > guide_tolerance) and apply_correction is True and mode == "auto":
                self.set_reference_frames()
                self.command.actor._status &= ~GuiderStatus.GUIDING
                self.command.actor._status |= GuiderStatus.ACQUIRING
                self.command.actor.status |= GuiderStatus.DRIFTING
                raise ValueError(
                    "Guide measured offset exceeds guide tolerance. "
                    "Skipping correction and reverting to acquisition."
                )

            # Calculate offset in motor axes.
            saz_diff_d, sel_diff_d = delta_radec2mot_axis(ra_ref, dec_ref, ra_p, dec_p)
            offset_motax = (saz_diff_d, sel_diff_d)

        self.command.info(
            measured_pointing={
                "frameno": frameno,
                "ra": ra_p,
                "dec": dec_p,
                "radec_offset": list(numpy.round(offset_radec, 3)),
                "motax_offset": list(numpy.round(offset_motax, 3)),
                "separation": sep,
                "mode": "guide" if self.use_reference_frames else "acquisition",
            }
        )

        # Calculate the correction.
        corr_radec = numpy.array(offset_radec)  # In RA/Dec
        corr_radec[0] = self.pid_ra(corr_radec[0])
        corr_radec[1] = self.pid_dec(corr_radec[1])

        corr_motax = numpy.array(offset_motax)  # In motor axes
        corr_motax[0] = self.pid_ra(corr_motax[0])
        corr_motax[1] = self.pid_dec(corr_motax[1])

        applied_radec = applied_motax = numpy.array([0.0, 0.0])
        try:
            if apply_correction:
                self.command.actor.status &= ~GuiderStatus.PROCESSING
                self.command.actor.status |= GuiderStatus.CORRECTING

                applied_radec, applied_motax = await self.offset_telescope(
                    *offset_motax,
                    use_motor_axes=use_motor_offsets,
                    max_correction=self.config.get("max_correction", 3600),
                )

        finally:
            self.command.actor.status &= ~GuiderStatus.CORRECTING

            self.command.info(
                correction_applied={
                    "frameno": frameno,
                    "radec_applied": list(numpy.round(applied_radec, 3)),
                    "motax_applied": list(numpy.round(applied_motax, 3)),
                }
            )

            # Should we start guiding?
            if (
                mode == "auto"
                and not self.use_reference_frames
                and sep < guide_tolerance
            ):
                self.command.warning("Guide tolerance reached. Starting to guide.")
                self.set_reference_frames(frameno, mf_wcs=wcs)
                self.command.actor._status &= ~GuiderStatus.ACQUIRING
                self.command.actor._status &= ~GuiderStatus.DRIFTING
                self.command.actor.status |= GuiderStatus.GUIDING

            asyncio.create_task(
                self.write_proc_file(
                    filenames=filenames,
                    wcs=wcs,
                    ra_p=ra_p,
                    dec_p=dec_p,
                    offset_radec=tuple(offset_radec),
                    corr_radec=tuple(corr_radec),
                    corr_motax=tuple(corr_motax),
                    sources=list(sources),
                    is_acquisition=is_acquisition,
                )
            )

    async def determine_pointing(
        self,
        filenames: list[str],
        pixel: tuple[float, float] | None = None,
        use_individual_images: bool = False,
    ) -> tuple[float, float, WCS]:
        """Returns the pointing of a telescope based on AG frames.

        Parameters
        ----------
        filenames
            A list of two AG frames (one for ``spec``) to be used to determine the
            pointing by calling `.solve_from_files`.
        pixel
            The ``(x,y)`` pixel of the master frame to use to determine the pointing.
            Default to the central pixel.
        use_individual_images
            Determine the coordinates of the central pixel of the master frame
            using a WCS generated from the individual frames, instead of trying
            to solve the master frame with astrometry.net.

        Returns
        -------
        pointing
            An tuple with the RA,Dec pointing of the telescope, in degrees.

        """

        telescope = self.telescope

        pixel = pixel or XZ_FULL_FRAME

        if not use_individual_images:
            solution: AstrometrySolution
            solution, _ = await run_in_executor(solve_from_files, filenames, telescope)
            wcs = solution.wcs
        else:
            wcs, solutions = await run_in_executor(
                wcs_from_single_cameras,
                filenames,
                telescope=telescope,
            )

            for camname in solutions:
                if solutions[camname].solved is False:
                    self.command.warning(f"Camera {camname} did not solve.")

        if wcs is None:
            raise ValueError(f"Cannot determine pointing for telescope {telescope}.")

        pointing: Any = wcs.pixel_to_world(*pixel)

        wcs = cast(WCS, wcs)

        return (numpy.round(pointing.ra.deg, 6), numpy.round(pointing.dec.deg, 6), wcs)

    def calculate_telescope_offset(
        self,
        pointing: tuple[float, float],
    ) -> tuple[tuple[float, float], tuple[float, float], float]:
        """Determines the offset to send to the telescope to acquire the field centre.

        Parameters
        ----------
        pointing
            The current pointing of the telescope, as determined
            by `.determine_pointing`.

        Returns
        -------
        offset
            A tuple of ra/dec offsets to acquire the desired field centre, in arcsec.
        angle
            The angle between pointing and field centre, in arcsec.

        """

        pra, pdec = pointing
        fra, fdec = self.field_centre

        # TODO: do this with proper spherical trigonometry! But won't matter much
        # if angle is small.

        mid_dec = (pdec + fdec) / 2

        ra_off: float = (fra - pra) * numpy.cos(numpy.radians(mid_dec))
        dec_off: float = fdec - pdec

        fdec_c = numpy.radians(90 - fdec)
        pdec_c = numpy.radians(90 - pdec)
        ra_diff_rad = numpy.radians(fra - pra)

        cos_dec = numpy.cos(fdec_c) * numpy.cos(pdec_c)
        sin_dec_cos_ra = numpy.sin(fdec_c) * numpy.sin(pdec_c) * numpy.cos(ra_diff_rad)
        cos_sep = cos_dec + sin_dec_cos_ra
        sep = numpy.degrees(numpy.arccos(cos_sep))

        ra_arcsec = numpy.round(ra_off * 3600, 3)
        dec_arcsec = numpy.round(dec_off * 3600, 3)
        sep_arcsec = numpy.round(sep * 3600, 3)

        saz_diff_d, sel_diff_d = delta_radec2mot_axis(fra, fdec, pra, pdec)

        return ((ra_arcsec, dec_arcsec), (saz_diff_d, sel_diff_d), sep_arcsec)

    async def offset_telescope(
        self,
        off0: float,
        off1: float,
        use_motor_axes: bool = False,
        max_correction: float | None = None,
    ):
        """Sends a correction offset to the telescope.

        Parameters
        ----------
        off0
            Offset in the first axis, in arcsec.
        off1
            Offset in the second axis, in arcsec.
        use_motor_axes
            Whether to apply the corrections as motor axes offsets.
        max_correction
            Maximum allowed correction. If any of the axes corrections is larger
            than this value an exception will be raised.

        Returns
        -------
        applied_radec
            The applied correction in RA and Dec in arcsec. If the correction
            is actually applied as motor offsets, this value is estimated from
            them.
        applied_motax
            The applied correction in motor axes in arcsec. Zero if the correction
            is applied as RA/Dec.

        """

        if numpy.any(numpy.abs([off0, off1]) > max_correction):
            raise ValueError("Requested correction is too big. Not applying it.")

        telescope = self.command.actor.telescope
        pwi = f"lvm.{telescope}.pwi"

        applied_radec = numpy.array([0.0, 0.0])
        applied_motax = numpy.array([0.0, 0.0])

        if use_motor_axes is False:
            cmd_str = f"offset --ra_add_arcsec {off0} --dec_add_arcsec {off1}"
            applied_radec = numpy.array([off0, off1])
        else:
            cmd_str = f"offset --axis0_add_arcsec {off0} --axis1_add_arcsec {off1}"
            applied_motax = numpy.array([off0, off1])
            # TODO: calculate applied_radec here. Maybe Tom has the
            # delta_mot_axis2radec equivalent, if not it should not be hard
            # to derive.

        cmd = await self.command.send_command(pwi, cmd_str)

        if cmd.status.did_fail:
            raise RuntimeError(f"Failed offsetting telescope {telescope}.")

        return applied_radec, applied_motax

    async def calculate_guide_offset(self, sources: list[pandas.DataFrame]):
        """Determines the guide offset by matching sources to reference images.

        Parameters
        ----------
        sources
            A Pandas DataFrame with the sources to match to the reference frame.
            Must contain a column ``camera`` with the camera associated with each
            source.

        Returns
        -------
        offset
            A tuple of RA and Dec offsets, in arcsec.
        separation
            The absolute separation between the reference frame and the
            new set of sources, in arcsec.

        """

        if self.reference_sources is None or self.reference_wcs is None:
            raise CriticalGuiderError("Missing reference frame data. Cannot guide.")

        offset, sep_arcsec, _ = await run_in_executor(
            calculate_guide_offset,
            pandas.concat(sources),
            self.telescope,
            self.reference_sources,
            self.reference_wcs,
        )

        # Rounding.
        offset = tuple(numpy.round(offset, 3))
        sep_arcsec = float(numpy.round(sep_arcsec, 3))

        # Typing.
        offset = cast(tuple[float, float], offset)
        sep_arcsec = cast(float, sep_arcsec)

        return offset, sep_arcsec

    async def write_proc_file(
        self,
        filenames: list[str],
        wcs: WCS | None,
        ra_p: float,
        dec_p: float,
        offset_radec: tuple[float, float],
        corr_radec: tuple[float, float],
        corr_motax: tuple[float, float],
        sources: list[pandas.DataFrame],
        is_acquisition=True,
    ):
        """Writes a ``proc-`` image with the astrometric info and applied corrections.

        Parameters
        ----------
        filenames
            A list of the paths of the individual AG frames used for the solution.
        wcs
            The master frame WCS. If ``is_acquisition=False``, this is the WCS of
            the master frame in the reference image.
        ra_p
            RA of the pointing
        dec_p
            Dec of the pointing.
        offset_radec
            Measured offset, in arcsec, in the RA/Dec direction.
        offset_motax
            Measured offset, in arcsec, in the motor axes direction.
        corr_radec
            Correction applied, in arcsec, in the RA/Dec direction.
        corr_motax
            Correction applied, in arcsec, in the motor axes direction.
        sources
            A list of the extracted sources in each AG frame.
        is_acquisition
            `True` if this was an acquisition step. `False` if guiding.

        """

        proc_path = get_proc_path(filenames[0])

        astro_hdu = fits.ImageHDU(name="ASTROMETRY")
        astro_hdr = astro_hdu.header

        astro_hdr["NAXIS"] = 2
        astro_hdr["ACQUISIT"] = (is_acquisition, "Acquisition or guiding?")

        for fn, file_ in enumerate(filenames):
            astro_hdr[f"FILE{fn}"] = (str(file_), f"AG frame {fn}")

        astro_hdr["RAFIELD"] = (self.field_centre[0], "[deg] Field RA")
        astro_hdr["DECFIELD"] = (self.field_centre[1], "[deg] Field Dec")
        astro_hdr["RAMEAS"] = (ra_p, "[deg] Measured RA position")
        astro_hdr["DECMEAS"] = (dec_p, "[deg] Measured Dec position")
        astro_hdr["OFFRAMEA"] = (offset_radec[0], "[arcsec] RA measured offset")
        astro_hdr["OFFDEMEA"] = (offset_radec[1], "[arcsec] Dec measured offset")
        astro_hdr["RACORR"] = (corr_radec[0], "[arcsec] RA applied correction")
        astro_hdr["DECORR"] = (corr_radec[1], "[arcsec] Dec applied correction")
        astro_hdr["AX0CORR"] = (corr_motax[0], "[arcsec] Motor axis 0 applied offset")
        astro_hdr["AX1CORR"] = (corr_motax[1], "[arcsec] Motor axis 1 applied offset")
        astro_hdr["RAKP"] = (self.pid_ra.Kp, "RA PID K term")
        astro_hdr["RAKI"] = (self.pid_ra.Ki, "RA PID I term")
        astro_hdr["RAKD"] = (self.pid_ra.Kd, "RA PID D term")
        astro_hdr["DECKP"] = (self.pid_dec.Kp, "Dec PID K term")
        astro_hdr["DECKI"] = (self.pid_dec.Ki, "Dec PID I term")
        astro_hdr["DECKD"] = (self.pid_dec.Kd, "Dec PID D term")

        if wcs is not None:
            astro_hdr += wcs.copy().to_header()

        proc_hdu = fits.HDUList([fits.PrimaryHDU(), astro_hdu])
        proc_hdu.append(
            fits.BinTableHDU(
                data=Table.from_pandas(pandas.concat(sources)),
                name="SOURCES",
            )
        )
        proc_hdu.writeto(str(proc_path))

        self.command.info(proc_path=str(proc_path))
        return proc_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-14
# @Filename: status.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

from lvmguider.actor import lvmguider_parser


if TYPE_CHECKING:
    from lvmguider.actor import GuiderCommand


__all__ = ["status"]


@lvmguider_parser.command()
async def status(command: GuiderCommand):
    """Outputs the status of the guider."""

    status = command.actor.status

    return command.finish(
        message={
            "status": f"0x{status.value:x}",
            "status_labels": ",".join(status.get_names()),
        }
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-10
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING

from clu.actor import AMQPActor

from lvmguider.cameras import Cameras
from lvmguider.maskbits import GuiderStatus


if TYPE_CHECKING:
    from lvmguider.guider import Guider

__all__ = ["LVMGuiderActor"]


class LVMGuiderActor(AMQPActor):
    """The ``lvmguider`` actor."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.model and self.model.schema:
            self.model.schema["additionalProperties"] = True

        self.telescope: str = self.config.get("telescope", self.name.split(".")[1])
        self.cameras = Cameras(self.telescope)

        self._status = GuiderStatus.IDLE

        self.guider: Guider | None = None
        self.guide_task: asyncio.Task | None = None

    @property
    def status(self):
        """Returns the guider status."""

        return self._status

    @status.setter
    def status(self, new_value: GuiderStatus):
        """Sets a new status and reports it to the users."""

        if new_value.value != self._status.value:
            self._status = new_value
            self.write(
                "d",
                {
                    "status": f"0x{self._status.value:x}",
                    "status_labels": ",".join(self._status.get_names()),
                },
                internal=True,
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-01-17
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
import sys

import click
from click_default_group import DefaultGroup

from sdsstools.daemonizer import DaemonGroup, cli_coro

from lvmguider import __version__
from lvmguider.actor.actor import LVMGuiderActor


@click.group(
    cls=DefaultGroup,
    default="actor",
    default_if_no_args=True,
    invoke_without_command=True,
)
@click.option(
    "--version",
    is_flag=True,
    help="Print version and exit.",
)
@click.option(
    "--config-file",
    "-c",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to the configuration file to use.",
)
@click.pass_context
def lvmguider(
    ctx: click.Context,
    config_file: str | None = None,
    version: bool = False,
):
    """HAL actor."""

    if version is True:
        click.echo(__version__)
        sys.exit(0)

    default_config_file = os.path.join(
        os.path.dirname(__file__),
        "etc/lvm.sci.guider.yml",
    )

    ctx.obj = {"config_file": config_file or default_config_file}


@lvmguider.group(cls=DaemonGroup, prog="lvmguider_actor", workdir=os.getcwd())
@click.pass_context
@cli_coro()
async def actor(ctx):
    """Runs the actor."""

    config_file = ctx.obj["config_file"]
    print("Configuration file", config_file)

    lvmguider_actor = LVMGuiderActor.from_config(config_file)

    await lvmguider_actor.start()
    await lvmguider_actor.run_forever()


def main():
    lvmguider(auto_envvar_prefix="LVMGUIDER")


if __name__ == "__main__":
    lvmguider()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-08
# @Filename: expose.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from lvmguider.actor import lvmguider_parser
from lvmguider.maskbits import GuiderStatus


if TYPE_CHECKING:
    from lvmguider.actor import GuiderCommand


__all__ = ["expose"]


@lvmguider_parser.command()
@click.option(
    "-t",
    "--exposure-time",
    type=float,
    default=5,
    help="Exposure time.",
)
@click.option(
    "-t",
    "--flavour",
    type=str,
    default="object",
    help="The type of exposure to take.",
)
@click.option(
    "--loop",
    is_flag=True,
    help="Loop exposures until guide stop is called.",
)
async def expose(
    command: GuiderCommand,
    exposure_time: float = 5.0,
    flavour="object",
    loop=False,
):
    """Exposes the cameras without guiding."""

    # Force the cameras to check the last image.
    command.actor.cameras.reset_seqno()

    if flavour not in ["object", "dark", "bias"]:
        return command.fail("Invalid flavour.")

    while True:
        await command.actor.cameras.expose(
            command,
            exposure_time=exposure_time,
            flavour=flavour,
            extract_sources=True,
        )
        if loop is False or (command.actor.status & GuiderStatus.STOPPING):
            break

    command.actor.status = GuiderStatus.IDLE

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-13
# @Filename: focus.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy
import pandas
from scipy.optimize import curve_fit  # type: ignore


if TYPE_CHECKING:
    from lvmguider.actor import GuiderCommand


def fit_hyperbola(
    x_arr: list[float],
    y_arr: list[float],
    y_err: list[float],
) -> tuple[float, float]:
    """Fit a hyperbola.

    Returns
    -------
        Minimum of hyperbola and its uncertainty.

    """

    # initial guess
    ic = numpy.argmin(y_arr)
    ix = numpy.argmax(y_arr)
    b = y_arr[ic]
    c = x_arr[ic]
    x = x_arr[ix]
    slope = numpy.abs((y_arr[ic] - y_arr[ix]) / (c - x))
    a = b / slope

    # initial values
    p0 = [a, b, c]

    # fit
    coeffs, cov = curve_fit(
        lambda xx, aa, bb, cc: bb * numpy.sqrt((xx - cc) ** 2 / aa**2 + 1.0),
        x_arr,
        y_arr,
        sigma=y_err,
        p0=p0,
    )

    # return result
    return coeffs[2], cov[2][2]


def fit_parabola(x_arr: list[float], y_arr: list[float], y_err: list[float]):
    """Fits a parabola to the data."""

    w = 1 / numpy.array(y_err)
    a, b, c = numpy.polyfit(x_arr, y_arr, 2, w=w, full=False)

    return (a, b, c)


class Focuser:
    """Performs focus on a telescope."""

    def __init__(
        self,
        command: GuiderCommand,
        guess: float | None = None,
        step_size: float = 0.5,
        steps: int = 7,
        exposure_time: float = 5.0,
    ):
        self.command = command
        self.initial_guess = guess
        self.step_size = step_size
        self.steps = int(steps)
        self.exposure_time = exposure_time

        if self.steps % 2 == 0:
            self.steps += 1

        self.telescope = self.command.actor.telescope
        self.cameras = self.command.actor.cameras
        self.foc_actor = f"lvm.{self.telescope}.foc"  # Focuser

    async def focus(self, max_cov: float = 5.0, curve="parabola"):
        """Performs the focus routine."""

        if self.initial_guess is None:
            cmd = await self.command.send_command(self.foc_actor, "getPosition")
            if cmd.status.did_fail:
                raise RuntimeError("Failed retrieving position from focuser.")
            self.initial_guess = cmd.replies.get("Position")
            self.command.debug(f"Using focuser position: {self.initial_guess} DT")

        focus_grid = numpy.arange(
            self.initial_guess - (self.steps // 2) * self.step_size,
            self.initial_guess + (self.steps // 2 + 1) * self.step_size,
            self.step_size,
        )

        if numpy.any(focus_grid <= 0):
            raise ValueError("Focus values out of range.")

        mean_fwhm = []
        source_list = []
        for focus_position in focus_grid:
            await self.goto_focus_position(focus_position)
            _, _, sources = await self.cameras.expose(
                self.command,
                exposure_time=self.exposure_time,
                extract_sources=True,
            )

            if sources is None or len(sources) == 0:
                self.command.warning(f"No sources detected at {focus_position} DT.")
                mean_fwhm.append(1e6)
                continue

            asources = pandas.concat(sources)
            if len(asources) == 0:
                self.command.warning(
                    f"No sources found for focus position {focus_position} DT. "
                    "Skipping this point."
                )
                continue

            asources["dt"] = focus_position

            fwhm = asources["xstd"].median()
            self.command.info(
                focus_point=dict(
                    focus=focus_position,
                    n_sources=len(asources),
                    fwhm=round(fwhm, 2),
                )
            )
            mean_fwhm.append(fwhm)

            source_list.append(asources)

        if len(source_list) < 3:
            raise ValueError("Insufficient number of focus points.")

        sources = pandas.concat(source_list)

        best = cov = fwhm_best = -999.0
        if curve == "hyperbola":
            best, cov = fit_hyperbola(
                sources.dt.tolist(),
                sources.xstd.tolist(),
                sources.xrms.tolist(),
            )

            if cov > max_cov:
                best_focus_idx = numpy.argmin(mean_fwhm)
                self.initial_guess = focus_grid[best_focus_idx]
                self.command.warning(
                    "Covariance not reached. Trying again "
                    f"with initial guess {self.initial_guess:.1f} DT."
                )
                await self.focus()

        elif curve == "parabola":
            a, b, c = fit_parabola(
                sources.dt.tolist(),
                sources.xstd.tolist(),
                sources.xrms.tolist(),
            )
            best = -b / 2 / a
            fwhm_best = a * best**2 + b * best + c
            cov = -999.0
            self.initial_guess = best

        self.command.info(
            best_focus=dict(
                focus=numpy.round(best, 1),
                fwhm=numpy.round(fwhm_best, 2),
                cov=numpy.round(cov, 2),
            )
        )
        await self.goto_focus_position(numpy.round(best, 2))

    async def goto_focus_position(self, focus_position: float):
        """Moves the focuser to a position."""

        cmd = await self.command.send_command(
            self.foc_actor,
            f"moveAbsolute {focus_position} DT",
        )
        if cmd.status.did_fail:
            raise RuntimeError(f"Failed reaching focus {focus_position:.1f} DT.")
        if cmd.replies.get("AtLimit") is True:
            raise RuntimeError("Hit a limit while focusing.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-18
# @Filename: transformations.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib
import warnings
from datetime import datetime

import numpy
import pandas
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.io import fits
from astropy.time import Time
from astropy.wcs import WCS
from astropy.wcs.utils import fit_wcs_from_points, pixel_to_skycoord
from scipy.spatial import KDTree

from lvmguider.astrometrynet import AstrometrySolution, astrometrynet_quick
from lvmguider.extraction import extract_marginal
from lvmguider.tools import get_proc_path


# Middle of an AG frame or full frame
XZ_FULL_FRAME = (2500.0, 1000.0)
XZ_AG_FRAME = (800.0, 550.0)


def ag_to_master_frame(
    camera: str,
    in_locs: numpy.ndarray,
    rot_shift: tuple | None = None,
    verbose=False,
):
    """Rotate and shift the star locations from an AG frame to the correct locations in
    the large "master frame" corresponding to the full focal plane.

    Adapted from Tom Herbst's code.

    Parameters
    ----------
    camera:
        Which AG camera: ``skyw-e``, ``skyw-w``, ``sci-c``, ``sci-e``, ``sci-w``,
        ``skye-e``, ``skye-w``, or ``spec-e``.
    in_locs
        Numpy 2D array of (x, y) star locations.
    rot_shift
        List of input ``[theta_deg, Sx, Sz]` to use. If `None`, use best fit values

    Returns
    -------
    rs_loc
        Array of corresponding star locations in "master frame".

    """

    camera = camera.lower()

    # Get best-fit rotation and shift, based on NonLinLSQ_Fit.py
    if rot_shift is None:  # Use Best-Fit values from Suff_AGC.docx
        # Best-fit rotations and shifts
        # Last one is Sci on-axis for 21 Feb 23 - see Doc
        met_data = {
            "skyw-w": [-89.85, -76.53, 430.64],
            "skyw-e": [90.16, 3512.90, 433.74],
            "sci-w": [-89.84, -28.26, 415.33],
            "sci-e": [89.98, 3512.74, 447.25],
            "spec-e": [89.60, 3193.11, 504.26],
            "skye-w": [-90.69, -79.54, 436.38],
            "skye-e": [89.71, 3476.00, 457.13],
            "sci-c": [-89.995, 1707.4, 426.62],
        }

        # Pull out correct best-fit rotation and shift
        bF = met_data[camera]
        th, Sx, Sz = (numpy.radians(bF[0]), bF[1], bF[2])

    else:
        # Use supplied values
        th, Sx, Sz = (numpy.radians(rot_shift[0]), rot_shift[1], rot_shift[2])

    if verbose:
        print("")
        print("ag_to_master_frame: rot_shift for ", camera, numpy.degrees(th), Sx, Sz)
        print("")

    # Rotation matrix
    M = numpy.array(([numpy.cos(th), -numpy.sin(th)], [numpy.sin(th), numpy.cos(th)]))

    # 2xnPts rotation offset
    rOff = numpy.tile(numpy.array([[800.0, 550.0]]).transpose(), (1, in_locs.shape[0]))

    # Make 2xnPts Sx,Sz offset matrix
    off = numpy.tile(numpy.array([[Sx, Sz]]).transpose(), (1, in_locs.shape[0]))

    # Shift to center, Rotate, shift back, and offset
    rsLoc = numpy.dot(M, in_locs.T - rOff) + rOff + off

    # Return calculated positions (need to transpose for standard layout)
    return (rsLoc.T, (th, Sx, Sz))


def solve_locs(
    locs: pandas.DataFrame,
    ra: float,
    dec: float,
    full_frame=True,
    index_paths: dict[int, str] | None = None,
    scales: dict[int, list[int]] | None = None,
    output_root: str | None = None,
    verbose: bool = False,
    raise_on_unsolved: bool = True,
    **astrometrynet_kwargs,
):
    """Use astrometry.net to solve field, given a series of star locations.

    Adapted from Tom Herbst's code.

    Parameters
    ----------
    locs
        A Pandas DataFrame of ``x``, ``y``, and ``flux`` star locations.
    ra
        The estimated RA of the field.
    dec
        The estimated Dec of the field.
    full_frame
        Whether this is a full "master frame" field, requiring a different set of
        index files.
    index_paths
        The paths to the index files to use. A dictionary of series number
        to path on disk, e.g.,
        ``{5200: "/data/astrometrynet/5200", 4100: "/data/astrometrynet/5100"}``.
    scales
        Index scales to use. Otherwise uses all index files. The format
        is a dictionary of series number to a list of scales, e.g.,
        ``{5200: [4, 5], 4100: [10, 11]}``.
    output_root
        The path where to write the astrometry.net files. If `None`, uses temporary
        files.
    verbose
        Output additional information.
    raise_on_unsolved
        Whether to raise an error if the field was not solved.
    astrometrynet_kwargs
        Arguments to pass directly to `.astrometrynet_quick`.

    Returns
    -------
    myWCS
        A WCS entity (or `None` if it fails) with appropriate location information.
    solveInfo
        A list containing information about the solve:
        solveInfo[0] - [raSlv,deSlv]  RA,Dec of central pixel
                 [1] - angSlv         Rotation of field
                 [2] - pxScl          Arcsec / pixel of solve

    """

    lower_bound = 0.8  # Pixel scale hints (arcsec/pixel)
    upper_bound = 1.2
    radius = 5  # Search radius in degrees

    if full_frame:
        midX, midZ = XZ_FULL_FRAME  # Middle of Master Frame
        index_paths_default = {
            5200: "/data/astrometrynet/5200",
            4100: "/data/astrometrynet/4100",
        }
    else:
        midX, midZ = XZ_AG_FRAME  # Middle of AG camera Frame
        index_paths_default = {5200: "/data/astrometrynet/5200"}

    index_paths = index_paths or index_paths_default
    scales = scales or {5200: [4, 5, 6]}

    locs = locs.copy()
    if full_frame:
        locs = locs.rename(columns={"x_master": "x", "y_master": "y"})

    solution = astrometrynet_quick(
        index_paths,
        locs,
        ra=ra,
        dec=dec,
        radius=radius,
        pixel_scale=(upper_bound + lower_bound) / 2,
        pixel_scale_factor_hi=upper_bound,
        pixel_scale_factor_lo=lower_bound,
        width=midX * 2,
        height=midZ * 2,
        scales=scales,
        raise_on_unsolved=raise_on_unsolved,
        output_root=output_root,
        **astrometrynet_kwargs,
    )

    if verbose:
        print("")
        print("Solution has match ? ", solution.solved)

    return solution


def solve_from_files(
    files: list[str],
    telescope: str,
    reextract_sources: bool = False,
    solve_locs_kwargs: dict | None = None,
):
    """Determines the telescope pointing from a set of AG frames.

    Parameters
    ----------
    files
        The AG FITS files to use determine the telescope pointing. Normally
        these are a pair of east/west camera frames, except for ``spec``.
    telescope
        The telescope to which these exposures are associated.
    reextract_sources
        Runs the source extraction algorithm again. If `False`, extraction is only
        done if a ``SOURCES`` extensions is not found in the files.
    solve_locs_kwargs
        A dictionary of kwargs to pass to `.solve_locs`.

    Returns
    -------
    wcs
        The wcs of the astrometric solution or `None` if no solution was found.
    locs
        The locations used to solve with astrometry.net. This is basically the
        extracted sources with two new ``x_master`` and ``y_master`` columns.

    """

    if len(files) == 0:
        raise ValueError("No files provided.")

    solve_locs_kwargs = solve_locs_kwargs if solve_locs_kwargs is not None else {}

    mf_sources: list[pandas.DataFrame] = []
    ra = 0.0
    dec = 0.0
    for file in files:
        header = fits.getheader(file)
        camname = header["CAMNAME"].lower()
        ra = header["RA"]
        dec = header["DEC"]

        if reextract_sources is False:
            try:
                sources = pandas.DataFrame(fits.getdata(file, "SOURCES"))
            except KeyError:
                warnings.warn("SOURCES ext not found. Extracting sources", UserWarning)
                return solve_from_files(
                    files,
                    telescope,
                    reextract_sources=True,
                    solve_locs_kwargs=solve_locs_kwargs,
                )
        else:
            hdul = fits.open(file)
            if "PROC" in hdul:
                data = hdul["PROC"].data
            else:
                data = hdul[0].data
            sources = extract_marginal(data)
            sources["camera"] = camname

        xy = sources[["x", "y"]].values

        camera = f"{telescope}-{camname[0]}"
        file_locs, _ = ag_to_master_frame(camera, xy)
        sources.loc[:, ["x_master", "y_master"]] = file_locs
        mf_sources.append(sources)

    locs = pandas.concat(mf_sources)

    if "output_root" not in solve_locs_kwargs:
        # Generate root path for astrometry files.
        proc_path = get_proc_path(files[0])
        dirname = proc_path.parent
        proc_base = proc_path.name.replace(".fits.gz", "").replace(".fits", "")
        output_root = str(dirname / "astrometry" / proc_base)
        solve_locs_kwargs["output_root"] = output_root

    solution = solve_locs(
        locs[["x_master", "y_master", "flux"]],
        ra=ra,
        dec=dec,
        full_frame=True,
        **solve_locs_kwargs,
    )

    return solution, locs


def radec2azel(raD, decD, lstD):
    """Returns the Azimuth and Elevation for the supplied RA, Dec, and sidereal time.

    From Tom Herbst and Florian Briegel.

    Parameters
    ----------
    ra,dec
        Right Ascension and Declination in decimal degrees.
    lst
        Local Sideral Time in decimal degrees.

    Returns
    -------
    az,el
        Azimuth and Elevation (Altitude) in decimal degrees

    """

    site = EarthLocation.of_site("Las Campanas Observatory")

    lat_r = numpy.radians(site.lat.deg)

    ra, dec, lst = (
        numpy.radians(raD),
        numpy.radians(decD),
        numpy.radians(lstD),
    )  # Convert to radians

    ha = lst - ra

    el = numpy.arcsin(
        numpy.sin(dec) * numpy.sin(lat_r)
        + numpy.cos(dec) * numpy.cos(lat_r) * numpy.cos(ha)
    )

    rat = (numpy.sin(dec) - numpy.sin(el) * numpy.sin(lat_r)) / (
        numpy.cos(el) * numpy.cos(lat_r)
    )  # Ratio - need to pin [-1,1]

    if rat < -1.0:  # Goes wonky if roundoff puts it outside [1,1]
        rat = -1.0

    if rat > 1.0:
        rat = 1.0

    if numpy.sin(ha) < 0.0:
        az = numpy.arccos(rat)
    else:
        az = 2.0 * numpy.pi - numpy.arccos(rat)

    return numpy.degrees(az), numpy.degrees(el)


def azel2sazsel(azD, elD):
    """Returns the siderostat coordinates (saz, Sel) for the supplied Az-El.

    From Tom Herbst and Florian Briegel.

    Parameters
    ----------
    az,el
        Azimuth and Elevation (Altitude) in decimal degrees

    Returns
    -------
    sazD,selD
        Siderostat angles in degrees

    """

    r90 = numpy.radians(90.0)  # 90 deg in radians
    az, el = numpy.radians(azD), numpy.radians(elD)  # Convert to radians
    SEl = numpy.arccos(numpy.cos(el) * numpy.cos(az)) - r90  # SEl in radians
    rat = numpy.sin(el) / numpy.cos(SEl)  # Ratio
    if azD < 180.0:
        SAz = r90 - numpy.arcsin(rat)  # saz in radians
    else:
        SAz = numpy.arcsin(rat) - r90

    return numpy.degrees(SAz), numpy.degrees(SEl)  # Return values in degrees


def delta_radec2mot_axis(ra_ref, dec_ref, ra_new, dec_new):
    """RA/Dec offset to motor axes.

    From Tom Herbst and Florian Briegel.

    """

    observing_location = EarthLocation.of_site("Las Campanas Observatory")
    observing_time = Time(datetime.utcnow(), scale="utc", location=observing_location)
    lst = observing_time.sidereal_time("mean")

    ref_az_d, ref_el_d = radec2azel(ra_ref, dec_ref, lst.deg)
    new_az_d, new_el_d = radec2azel(ra_new, dec_new, lst.deg)

    ref_saz_d, ref_sel_d = azel2sazsel(ref_az_d, ref_el_d)
    new_saz_d, new_sel_d = azel2sazsel(new_az_d, new_el_d)

    saz_diff_d = ref_saz_d - new_saz_d
    sel_diff_d = ref_sel_d - new_sel_d

    # To arcsec
    saz_diff_arcsec = saz_diff_d * -3600.0
    sel_diff_arcsec = sel_diff_d * -3600.0

    return saz_diff_arcsec, sel_diff_arcsec


def wcs_from_single_cameras(
    files: list[str | pathlib.Path],
    telescope: str | None = None,
    method: str = "astropy",
    reextract_sources: bool = False,
    solve_locs_kwargs: dict | None = None,
) -> tuple[WCS, dict[str, AstrometrySolution]]:
    """Determines the telescope pointing using individual AG frames.

    The main difference between this function and `.solve_from_files` is
    that here we solve each AG frame independently with astrometry.net and
    then use those solutions to generate a master frame WCS.

    Parameters
    ----------
    files
        The AG FITS files to use determine the telescope pointing. Normally
        these are a pair of east/west camera frames, except for ``spec``.
    telescope
        The telescope to which these exposures are associated. If `None`, it
        is determine from the header.
    method
        Either ``'astropy'`` or ``'tom'``. The method used to generate the master
        WCS from the independent solutions.
    reextract_sources
        Runs the source extraction algorithm again. If `False`, extraction is only
        done if a ``SOURCES`` extensions is not found in the files.
    solve_locs_kwargs
        A dictionary of kwargs to pass to `.solve_locs`.

    Returns
    -------
    wcs
        The wcs of the master frame astrometric solution.
    solutions
        A dictionary of the astrometric solutions for each camera.

    """

    if len(files) == 0:
        raise ValueError("No files provided.")

    solve_locs_kwargs = solve_locs_kwargs if solve_locs_kwargs is not None else {}
    solutions: dict[str, AstrometrySolution] = {}

    for file in files:
        file = pathlib.Path(file).absolute()
        header = fits.getheader(file)
        camname = header["CAMNAME"].lower()
        ra = header["RA"]
        dec = header["DEC"]
        telescope = telescope or header["TELESCOP"].lower()

        if reextract_sources is False:
            try:
                sources = pandas.DataFrame(fits.getdata(file, "SOURCES"))
            except KeyError:
                warnings.warn("SOURCES ext not found. Extracting sources", UserWarning)
                return wcs_from_single_cameras(
                    files,
                    telescope=telescope,
                    reextract_sources=True,
                    solve_locs_kwargs=solve_locs_kwargs,
                )
        else:
            hdul = fits.open(file)
            if "PROC" in hdul:
                data = hdul["PROC"].data
            else:
                data = hdul[0].data
            sources = extract_marginal(data)
            sources["camera"] = camname

        solve_locs_kwargs_cam = solve_locs_kwargs.copy()
        if "output_root" not in solve_locs_kwargs_cam:
            # Generate root path for astrometry files.
            basename = file.name.replace(".fits.gz", "").replace(".fits", "")
            output_root = str(file.parent / "astrometry" / basename)
            solve_locs_kwargs_cam["output_root"] = output_root

        camera_solution = solve_locs(
            sources.loc[:, ["x", "y", "flux"]],
            ra,
            dec,
            full_frame=False,
            raise_on_unsolved=False,
            **solve_locs_kwargs_cam,
        )

        if camera_solution.solved and camera_solution.stars is not None:
            camera = f"{telescope}-{camname[0]}"
            xy = camera_solution.stars.loc[:, ["field_x", "field_y"]].to_numpy()
            mf_locs, _ = ag_to_master_frame(camera, xy)
            camera_solution.stars.loc[:, ["x_master", "y_master"]] = mf_locs

        solutions[camname] = camera_solution

    nsolved = len([sol for sol in solutions.values() if sol.solved])
    if nsolved == 0:
        raise ValueError("No solutions found.")

    # Check for repeat images, in which East and West are almost identical. In
    # this case the coordinates of their central pixels will also be very close,
    # instead of being about a degree apart.
    wcs_east = solutions["east"].wcs if "east" in solutions else None
    wcs_west = solutions["west"].wcs if "west" in solutions else None
    if telescope != "spec" and wcs_east is not None and wcs_west is not None:
        east_cen = wcs_east.pixel_to_world(*XZ_AG_FRAME)
        west_cen = wcs_west.pixel_to_world(*XZ_AG_FRAME)
        ew_separation = east_cen.separation(west_cen).arcsec
        if ew_separation < 3000:
            raise ValueError("Found potential double image.")

    # Build a master frame WCS from the individual WCS solutions. We implement two
    # methods. Astropy uses the fit_wcs_from_points routine, where we use the
    # coordinates of the stars identified by astrometry.net and their corresponding
    # xy pixels on the master frame (as determined above with ag_to_master_frame).
    # Tom's method manually builds the WCS using the individual WCS solutions.
    if method == "astropy":
        all_stars = pandas.concat(
            [ss.stars for ss in solutions.values() if ss.stars is not None]
        )
        if len(all_stars) == 0:
            raise ValueError("No solved fields found.")

        skycoords = SkyCoord(
            ra=all_stars.index_ra,
            dec=all_stars.index_dec,
            unit="deg",
            frame="icrs",
        )
        wcs = fit_wcs_from_points((all_stars.x_master, all_stars.y_master), skycoords)

    elif method == "tom":
        if len(solutions) != 2:
            raise ValueError("Tom's method requires two cameras.")

        assert wcs_west is not None and wcs_east is not None, "Found unsolved fields."

        west_sky = wcs_west.pixel_to_world(*XZ_AG_FRAME)

        # RA, Dec of central pixel of West AG Camera
        CRVAL = numpy.array([west_sky.ra.deg, west_sky.dec.deg])  # type:ignore
        # Pixel location in MF of central px of W camera
        CRPIX = ag_to_master_frame(f"{telescope}-w", numpy.array([XZ_AG_FRAME]))[0][0]

        deg_per_pix = 1.0089 / 3600.0  # Pixel scale in degrees / pixel
        CDELT = numpy.array([-1.0 * deg_per_pix, deg_per_pix])  # CDELT entity for WCS

        # Derive position angle of field
        # West point in AG frame
        west_PA = master_frame_to_ag(f"{telescope}-w", numpy.array([[700.0, 1000.0]]))
        # and East point
        east_PA = master_frame_to_ag(f"{telescope}-e", numpy.array([[4300.0, 1000.0]]))

        # Get sky coords of these pixels
        wPA_sky = pixel_to_skycoord(west_PA[0][0], west_PA[0][1], wcs_west)
        ePA_sky = pixel_to_skycoord(east_PA[0][0], east_PA[0][1], wcs_east)

        # This is the angle we need
        west_pa = wPA_sky.position_angle(ePA_sky).degree  # type: ignore
        PA = numpy.radians((west_pa + 90.0) % 360)

        # Rotation matrix
        PC = numpy.array(
            [
                [numpy.cos(PA), -numpy.sin(PA)],
                [numpy.sin(PA), numpy.cos(PA)],
            ]
        )

        # Assemble WCS
        wcs = WCS(naxis=2)  # Create new WCS entity
        wcs.wcs.ctype = ["RA---TAN", "DEC--TAN"]  # Set CTYPE
        wcs.wcs.crpix = CRPIX  # Reference pixel indices
        wcs.wcs.crval = CRVAL  # Reference pixel RA/Dec
        wcs.wcs.cdelt = CDELT  # Pixel scale in deg/px
        wcs.wcs.pc = PC  # Rotation matrix

    else:
        raise ValueError("Invalid method.")

    return wcs, solutions


def master_frame_to_ag(camera: str, locs: numpy.ndarray):
    """Transforms the star locations from the master frame
    to the internal pixel locations in the AG frame.

    Adapted from Tom Herbst.

    Parameters
    ----------
    camera:
        Which AG camera: ``skyw-e``, ``skyw-w``, ``sci-c``, ``sci-e``, ``sci-w``,
        ``skye-e``, ``skye-w``, or ``spec-e``.
    locs
        Numpy 2D array of (x, y) star locations.
    Returns
    -------
    ag_loc
        Array of corresponding star locations in the AG frame.

    """

    camera = camera.lower()

    met_data = {
        "skyw-w": [-89.85, -76.53, 430.64],
        "skyw-e": [90.16, 3512.90, 433.74],
        "sci-w": [-89.84, -28.26, 415.33],
        "sci-e": [89.98, 3512.74, 447.25],
        "spec-e": [89.60, 3193.11, 504.26],
        "skye-w": [-90.69, -79.54, 436.38],
        "skye-e": [89.71, 3476.00, 457.13],
        "sci-c": [-89.995, 1707.4, 426.62],
    }

    # Pull out correct best-fit rotation and shift
    bF = met_data[camera]
    th, Sx, Sz = (numpy.radians(bF[0]), bF[1], bF[2])

    # Rotation matrix
    M = numpy.array(([numpy.cos(th), -numpy.sin(th)], [numpy.sin(th), numpy.cos(th)]))

    # 2xnPts rotation offset
    rOff = numpy.tile(numpy.array([[800.0, 550.0]]).transpose(), (1, locs.shape[0]))

    # Make 2xnPts Sx,Sz offset matrix
    off = numpy.tile(numpy.array([[Sx, Sz]]).transpose(), (1, locs.shape[0]))

    # Transform MF --> AG. See Single_Sensor_Solve doc
    rsLoc = numpy.dot(M, (locs.T - rOff - off)) + rOff

    # Return calculated positions (need to transpose for standard layout)
    return rsLoc.T


def calculate_guide_offset(
    sources: pandas.DataFrame,
    telescope: str,
    reference_sources: pandas.DataFrame,
    reference_wcs: WCS,
    max_separation: float = 5,
) -> tuple[tuple[float, float], float, pandas.DataFrame]:
    """Determines the guide offset by matching sources to reference images.

    Parameters
    ----------
    sources
        A list of sources extracted from the AG frames. Must contain a column
        ``camera`` with the camera associated with each source.
    telescope
        The telescope associated with these images.
    reference_sources
        As ``sources``, a list of extracted detections from the reference frames.
    reference_wcs
        The WCS of the master frame corresponding to the reference images.
    max_separation
        Maximum separation between reference and test sources, in pixels.

    Returns
    -------
    offset
        A tuple of RA and Dec offsets, in arcsec. This is the offset to
        go from current pointing to the reference pointing.
    separation
        The absolute separation between the reference frame and the
        new set of sources, in arcsec.
    matches
        A Pandas data frame with the matches between test and reference sources.

    """

    sources["camera"] = sources["camera"].astype(str)
    reference_sources["camera"] = reference_sources["camera"].astype(str)

    matches: pandas.DataFrame | None = None

    cameras = sources["camera"].unique()
    for camera in cameras:
        cam_ref_sources = reference_sources.loc[reference_sources["camera"] == camera]
        cam_sources = sources.loc[sources["camera"] == camera]

        if len(cam_ref_sources) == 0 or len(cam_sources) == 0:
            continue

        # Create a KDTree with the reference pixels.
        tree = KDTree(cam_ref_sources.loc[:, ["x", "y"]].to_numpy())

        # Find nearest neighbours
        dd, ii = tree.query(cam_sources.loc[:, ["x", "y"]].to_numpy())

        # Select valid matches.
        valid = dd < max_separation

        # Get the valid xy in the test and reference source lists.
        cam_matches = cam_sources.loc[valid, ["x", "y"]]
        cam_ref_matches = cam_ref_sources.iloc[ii[valid]]
        cam_ref_matches = cam_ref_matches.loc[:, ["x", "y"]]

        # Calculate MF coordinates for test and references pixels.
        camname = f"{telescope}-{camera[0]}"
        cam_matches_mf, _ = ag_to_master_frame(camname, cam_matches.to_numpy())
        cam_ref_matches_mf, _ = ag_to_master_frame(camname, cam_ref_matches.to_numpy())

        # Build a DF with the MF coordinates. Add the camera name and concatenate
        # to build a main DF with all the matches for both cameras.

        matches_array = numpy.hstack(
            (
                cam_matches,
                cam_ref_matches,
                cam_matches_mf,
                cam_ref_matches_mf,
            )
        )

        matches_df = pandas.DataFrame(
            matches_array,
            columns=[
                "x",
                "y",
                "xref",
                "yref",
                "x_mf",
                "y_mf",
                "xref_mf",
                "yref_mf",
            ],
        )
        matches_df["camera"] = camname

        if matches is None:
            matches = matches_df
        else:
            matches = pandas.concat([matches, matches_df])

    if matches is None:
        raise ValueError("No matches found.")

    # Determine offsets in xy.
    offset_x = matches.x_mf - matches.xref_mf
    offset_y = matches.y_mf - matches.yref_mf

    # Rotate to align with RA/Dec
    PC = reference_wcs.wcs.pc if reference_wcs.wcs.has_pc() else reference_wcs.wcs.cd
    offset_ra_d, offset_dec_d = PC @ numpy.array([offset_x, offset_y])

    # To arcsec.
    offset_ra_arcsec = numpy.mean(offset_ra_d * 3600)
    offset_dec_arcsec = numpy.mean(offset_dec_d * 3600)

    # Calculate separation. This is somewhat approximate but good enough
    # for small angles.
    sep_arcsec = numpy.sqrt(offset_ra_arcsec**2 + offset_dec_arcsec**2)

    return (offset_ra_arcsec, offset_dec_arcsec), sep_arcsec, matches



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-13
# @Filename: camera.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import os
import pathlib
import re
import warnings

from typing import TYPE_CHECKING

import numpy
import pandas
from astropy.io import fits
from astropy.table import Table

from sdsstools.time import get_sjd

from lvmguider.extraction import extract_marginal
from lvmguider.maskbits import GuiderStatus
from lvmguider.tools import run_in_executor


if TYPE_CHECKING:
    from lvmguider.actor import GuiderCommand


__all__ = ["Cameras"]


class Cameras:
    """Exposures and solves the telescope cameras."""

    def __init__(self, telescope: str):
        self.telescope = telescope

        self.agcam = f"lvm.{telescope}.agcam"

        self.pwi = f"lvm.{telescope}.pwi"
        self.km = f"lvm.{telescope}.km"
        self.foc = f"lvm.{telescope}.foc"

        self.sjd = get_sjd("LCO")
        self.last_seqno = -1

        self.dark_file: dict[str, str] = {}

    async def expose(
        self,
        command: GuiderCommand,
        exposure_time: float = 5.0,
        flavour: str = "object",
        extract_sources: bool = False,
        nretries: int = 3,
    ) -> tuple[list[str], int, list[pandas.DataFrame] | None]:
        """Exposes the cameras and returns the filenames."""

        command.actor._status &= ~GuiderStatus.IDLE
        command.actor.status |= GuiderStatus.EXPOSING

        # Update the status of the telescope and get the focuser position.
        await command.send_command(self.pwi, "status", internal=True)

        try:
            focus_cmd = await command.send_command(self.foc, "status", internal=True)
            focus_position = focus_cmd.replies.get("Position")
        except Exception as err:
            warnings.warn(f"Failed getting {self.telescope} focuser position: {err}")
            focus_position = -999

        if self.telescope != "spec":
            await command.send_command(self.pwi, "status", internal=True)

        next_seqno = self.get_next_seqno()

        command.debug(f"Taking agcam exposure {self.telescope}-{next_seqno}.")
        cmd = await command.send_command(
            self.agcam,
            f"expose -n {next_seqno} --{flavour} {exposure_time}",
        )

        if cmd.status.did_fail:
            command.actor._status &= ~GuiderStatus.EXPOSING
            command.actor._status |= GuiderStatus.FAILED
            command.actor.status |= GuiderStatus.IDLE
            raise RuntimeError("Failed while exposing cameras.")
        else:
            self.last_seqno = next_seqno

        filenames: set[str] = set()
        for reply in cmd.replies:
            for cam_name in ["east", "west"]:
                if cam_name in reply.message:
                    if reply.message[cam_name].get("state", None) == "written":
                        filename = reply.message[cam_name]["filename"]
                        filenames.add(filename)
                        if flavour == "dark":
                            self._write_dark_info(cam_name, filename)

        if len(filenames) == 0:
            raise ValueError("Exposure did not produce any images.")

        if any([self.is_shifted(filename) for filename in filenames]):
            if nretries > 0:
                command.warning("Found shifted images. Retrying.")
                return await self.expose(
                    command,
                    exposure_time=exposure_time,
                    flavour=flavour,
                    extract_sources=extract_sources,
                    nretries=nretries - 1,
                )
            else:
                raise RuntimeError("Run out of retries. Exposing failed.")

        # Create a new extension with the dark-subtracted image.
        for fn in filenames:
            with fits.open(fn, mode="update") as hdul:
                data = hdul[0].data.astype(numpy.float32)
                exptime = hdul[0].header["EXPTIME"]
                camname = hdul[0].header["CAMNAME"].lower()

                dark_file = self._get_dark_frame(fn, camname)
                if dark_file is None:
                    command.warning(f"No dark frame found for camera {camname}.")
                    continue

                dark_data = fits.getdata(dark_file).astype(numpy.float32)
                dark_exptime = fits.getheader(dark_file)["EXPTIME"]

                data_sub = data - (dark_data / dark_exptime) * exptime

                proc_header = hdul[0].header.copy()
                proc_header["DARKFILE"] = dark_file
                hdul.append(
                    fits.ImageHDU(
                        data=data_sub,
                        header=proc_header,
                        name="PROC",
                    )
                )

        sources = []
        if flavour == "object" and extract_sources:
            try:
                sources = await asyncio.gather(
                    *[self.extract_sources(fn) for fn in filenames]
                )
            except Exception as err:
                command.warning(f"Failed extracting sources: {err}")
                extract_sources = False
            else:
                for ifn, fn in enumerate(filenames):
                    with fits.open(fn, mode="update") as hdul:
                        camname = hdul[0].header["CAMNAME"].lower()
                        isources = sources[ifn]
                        isources["camera"] = camname
                        hdul.append(
                            fits.BinTableHDU(
                                data=Table.from_pandas(isources),
                                name="SOURCES",
                            )
                        )

        if len(sources) > 0:
            all_sources = pandas.concat(sources)
            fwhm = numpy.median(all_sources["xstd"]) if len(all_sources) > 0 else None
        else:
            all_sources = []
            fwhm = None

        command.info(
            frame={
                "seqno": next_seqno,
                "filenames": list(filenames),
                "flavour": flavour,
                "n_sources": len(all_sources),
                "focus_position": round(focus_position, 1),
                "fwhm": numpy.round(fwhm, 3) if fwhm else fwhm,
            }
        )

        command.actor.status &= ~GuiderStatus.EXPOSING

        if not command.actor.status & GuiderStatus.NON_IDLE:
            command.actor.status |= GuiderStatus.IDLE

        return (list(filenames), next_seqno, list(sources) if extract_sources else None)

    async def extract_sources(self, filename: str):
        """Extracts sources from a file."""

        hdus = fits.open(filename)
        if "PROC" in hdus:
            data = hdus["PROC"].data
        else:
            data = hdus[0].data

        return await run_in_executor(
            extract_marginal,
            data,
            box_size=31,
            threshold=3.0,
            max_detections=50,
            sextractor_quick_options={"minarea": 5},
        )

    def reset_seqno(self):
        """Resets the seqno.

        This forces a full check of the files next time that
        `.get_next_seqno` is called.

        """

        self.last_seqno = -1

    def get_next_seqno(self):
        """Determines the next exposure sequence number."""

        if get_sjd("LCO") == self.sjd and self.last_seqno >= 0:
            return self.last_seqno + 1

        self.sjd = get_sjd("LCO")

        path = pathlib.Path(f"/data/agcam/{self.sjd}")
        if not path.exists():
            self.last_seqno = 0
            return self.last_seqno + 1

        files = path.glob(f"*{self.telescope}*")
        seqnos = []
        for fn in files:
            se = re.search(r"\_([0-9]+)", str(fn))
            if not se:
                continue
            seqnos.append(int(se.group(1)))

        if len(seqnos) == 0:
            self.last_seqno = 0
        else:
            self.last_seqno = max(seqnos)

        return self.last_seqno + 1

    def is_shifted(self, filename: str | pathlib.Path | numpy.ndarray):
        """Determines if an image is "shifted"."""

        data: numpy.ndarray
        if isinstance(filename, (str, pathlib.Path)):
            data = fits.getdata(str(filename))
        else:
            data = filename

        # If more than 10% of the pixels have values >32,000 this is probably
        # a shifted image.
        npix = data.size
        if (data > 32000).sum() / npix > 0.1:
            return True

        # Another case of a bad image.
        if data.sum() == 0:
            return True

        return False

    def _get_dark_frame(self, filename: str, cam_name: str):
        """Gets the path to the dark frame."""

        dark_file = self.dark_file.get(cam_name, None)
        if dark_file is not None:
            return dark_file

        dirname = os.path.dirname(filename)
        path = os.path.join(dirname, f"lvm.{self.telescope}.agcam.{cam_name}.dark.dat")

        if os.path.exists(path):
            dark_file = open(path, "r").read().strip()
            self.dark_file[cam_name] = dark_file
            return dark_file

        return None

    def _write_dark_info(self, cam_name: str, filename: str):
        """Writes a file with the current dark frame."""

        self.dark_file[cam_name] = filename

        dirname = os.path.dirname(filename)
        path = os.path.join(dirname, f"lvm.{self.telescope}.agcam.{cam_name}.dark.dat")
        with open(path, "w") as fd:
            print(str(filename), file=fd)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-13
# @Filename: focus.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from lvmguider.actor import lvmguider_parser
from lvmguider.focus import Focuser


if TYPE_CHECKING:
    from lvmguider.actor import GuiderCommand


__all__ = ["focus"]


@lvmguider_parser.command()
@click.option(
    "-g",
    "--guess",
    type=float,
    help="Initial focus guess, in DT. If not provied, uses current focuser.",
)
@click.option(
    "-s",
    "--step-size",
    type=float,
    default=0.5,
    help="Step size, in DT.",
)
@click.option(
    "-n",
    "--steps",
    type=float,
    default=7,
    help="Number of steps (will be rounded to the closest odd number).",
)
@click.option(
    "-t",
    "--exposure-time",
    type=float,
    default=10,
    help="Exposure time in seconds.",
)
async def focus(
    command: GuiderCommand,
    guess: float | None = None,
    step_size: float = 0.5,
    steps: int = 7,
    exposure_time: float = 10,
):
    """Performs a focus sweep."""

    # Force the cameras to check the last image.
    command.actor.cameras.reset_seqno()

    focuser = Focuser(
        command,
        guess=guess,
        step_size=step_size,
        steps=steps,
        exposure_time=exposure_time,
    )

    try:
        await focuser.focus()
    except Exception as err:
        return command.fail(err)

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-13
# @Filename: tools.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import concurrent.futures
import pathlib
import re
import shutil
import warnings
from functools import partial

import pandas
from astropy.io import fits
from astropy.table import Table


async def run_in_executor(fn, *args, catch_warnings=False, executor="thread", **kwargs):
    """Runs a function in an executor.

    In addition to streamlining the use of the executor, this function
    catches any warning issued during the execution and reissues them
    after the executor is done. This is important when using the
    actor log handler since inside the executor there is no loop that
    CLU can use to output the warnings.

    In general, note that the function must not try to do anything with
    the actor since they run on different loops.

    """

    fn = partial(fn, *args, **kwargs)

    if executor == "thread":
        executor = concurrent.futures.ThreadPoolExecutor
    elif executor == "process":
        executor = concurrent.futures.ProcessPoolExecutor
    else:
        raise ValueError("Invalid executor name.")

    if catch_warnings:
        with warnings.catch_warnings(record=True) as records:
            with executor() as pool:
                result = await asyncio.get_event_loop().run_in_executor(pool, fn)

        for ww in records:
            warnings.warn(ww.message, ww.category)

    else:
        with executor() as pool:
            result = await asyncio.get_running_loop().run_in_executor(pool, fn)

    return result


def create_summary_table(
    path: str | pathlib.Path = ".",
    pattern: str = "lvm.*.agcam*.fits*",
) -> Table:
    """Returns a table with a summary of agcam information for the files in a directory.

    Parameters
    ----------
    path
        The path in which to search for file.
    pattern
        A pattern to select files. Defaults to selecting all raw frames.

    Returns
    -------
    table
        A table of frame data.

    """

    path = pathlib.Path(path)
    files = path.glob(pattern)

    columns = [
        "frameno",
        "telescope",
        "camera",
        "filename",
        "date-obs",
        "exptime",
        "imagetype",
        "ra",
        "dec",
        "nsources",
    ]
    data = []
    for file_ in files:
        file_match = re.match(r"lvm\.([a-z]+)\..+?\_([0-9]+)\.fits", file_.name)

        telescope = file_match.group(1) if file_match else ""
        frameno = int(file_match.group(2)) if file_match else -1

        hdus = fits.open(file_)
        if "SOURCES" in hdus:
            nsources = len(hdus["SOURCES"].data)
        else:
            nsources = -1

        header = hdus[0].header

        data.append(
            (
                frameno,
                telescope,
                header["CAMNAME"],
                file_.name,
                header["DATE-OBS"],
                header["EXPTIME"],
                header["IMAGETYP"],
                header["RA"],
                header["DEC"],
                nsources,
            )
        )

    tt = Table(rows=data, names=columns)
    tt.sort(["frameno", "telescope", "camera"])

    return tt


def get_proc_path(filename: str | pathlib.Path):
    """Returns the proc- path for a given file."""

    path = pathlib.Path(filename).absolute()
    dirname = path.parent
    basename = path.name.replace(".east", "").replace(".west", "")
    proc_path = dirname / ("proc-" + basename)

    return proc_path


def reprocess_proc_image(
    filename: str | pathlib.Path,
    telescope: str,
    output_path: str | pathlib.Path,
    keep_previous_wcs: bool = True,
    solve_individual_cameras: bool = True,
    generate_astrometrynet_outputs: bool = False,
    index_paths: dict[int, str] | None = None,
    scales: dict[int, list[int]] | None = None,
):
    """Reprocesses a proc- file."""

    from lvmguider.astrometrynet import astrometrynet_quick
    from lvmguider.transformations import ag_to_master_frame, solve_locs

    proc_orig = pathlib.Path(filename).absolute()
    output_path = pathlib.Path(output_path).absolute()

    hdus = fits.open(str(proc_orig))
    header = hdus[1].header

    ra = header["RAFIELD"]
    dec = header["DECFIELD"]

    if "SOURCES" not in hdus:
        warnings.warn("No SOURCES found.", UserWarning)
        return None

    sources = pandas.DataFrame(hdus["SOURCES"].data)
    sources.loc[:, "x_master"] = 0.0
    sources.loc[:, "y_master"] = 0.0

    output_path.mkdir(parents=True, exist_ok=True)
    proc_new = output_path / proc_orig.name

    shutil.copyfile(proc_orig, proc_new)

    proc_new = output_path / proc_orig.name
    new_sources_cam = []
    for camname in ["east", "west"]:
        sources_cam = sources.loc[sources.camera == camname]
        xy = sources_cam[["x", "y"]].values

        camera = f"{telescope}-{camname[0]}"
        file_locs, _ = ag_to_master_frame(camera, xy)
        sources_cam.loc[:, ["x_master", "y_master"]] = file_locs
        new_sources_cam.append(sources_cam)

    new_sources = pandas.concat(new_sources_cam)
    output_root = str(proc_new).replace(".fits.gz", "").replace(".fits", "")

    xyls = new_sources.loc[:, ["x_master", "y_master", "flux"]]
    xyls.rename(columns={"x_master": "x", "y_master": "y"}, inplace=True)

    try:
        solution = solve_locs(
            xyls,
            ra=ra,
            dec=dec,
            full_frame=True,
            index_paths=index_paths,
            scales=scales,
            output_root=output_root if generate_astrometrynet_outputs else None,
        )
        new_wcs = solution.wcs
    except RuntimeError:
        print(f"[WARNING]: Failed solving {filename}.")
        solution = None
        new_wcs = None

    with fits.open(str(proc_new), "update") as hdul:
        del hdul["SOURCES"]
        new_sources_t = Table.from_pandas(new_sources)
        hdul.append(fits.BinTableHDU(data=new_sources_t, name="SOURCES"))

        if new_wcs is None and keep_previous_wcs:
            hdul.append(fits.ImageHDU(name="REPROC"))
        elif new_wcs:
            if keep_previous_wcs:
                new_hdu = fits.ImageHDU(name="REPROC")
                new_hdu.header += new_wcs.to_header(relax=True)
                hdul.append(new_hdu)
            else:
                orig_hdu = hdul[1]
                orig_hdu.header += new_wcs.to_header(relax=True)

    if solve_individual_cameras:
        for camname in ["east", "west"]:
            sources_cam = new_sources.loc[new_sources.camera == camname]

            solution = astrometrynet_quick(
                {5200: "/data/astrometrynet/5200"},
                sources_cam,
                ra=ra,
                dec=dec,
                radius=5.0,
                pixel_scale=1.0,
                pixel_scale_factor_hi=1.2,
                pixel_scale_factor_lo=0.8,
                scales={5200: [5, 6]},
                plot=False,
                output_root=(output_root + f"_{camname}") if output_root else None,
            )

            with fits.open(str(proc_new), "update") as hdul:
                new_hdu = fits.ImageHDU(name=camname.upper())
                if solution.wcs:
                    new_hdu.header += solution.wcs.to_header(relax=True)
                    new_hdu.header["SOLVED"] = True
                else:
                    print(f"[WARNING]: Failed solving {filename} ({camname}).")
                    new_hdu.header["SOLVED"] = False
                hdul.append(new_hdu)

    return solution


def reprocess_files(
    filenames: list[str],
    telescope: str,
    proc_output_path: str | pathlib.Path,
    solve_individual_cameras: bool = True,
    generate_astrometrynet_outputs: bool = False,
    index_paths: dict[int, str] | None = None,
    scales: dict[int, list[int]] | None = None,
    **kwargs,
):
    """Reprocesses raw files.

    This basically calls `.solve_from_files` but writes the proc- file
    to a new location and allows to specify indexes and scales.

    """

    from lvmguider.astrometrynet import astrometrynet_quick
    from lvmguider.transformations import solve_from_files

    proc_output_path = pathlib.Path(proc_output_path)
    parent = proc_output_path.parent
    parent.mkdir(parents=True, exist_ok=True)

    output_root = (
        str(proc_output_path).replace(".fits.gz", "").replace(".fits", "")
        if generate_astrometrynet_outputs
        else None
    )

    solution, new_sources = solve_from_files(
        filenames,
        telescope,
        reextract_sources=True,
        solve_locs_kwargs={
            "index_paths": index_paths,
            "output_root": output_root,
            "scales": scales,
            **kwargs,
        },
    )
    assert solution.wcs

    new_sources_t = Table.from_pandas(new_sources)
    hdul = fits.HDUList(
        [
            fits.PrimaryHDU(header=solution.wcs.to_header(relax=True)),
            fits.BinTableHDU(data=new_sources_t, name="SOURCES"),
        ]
    )

    if solve_individual_cameras:
        header = fits.getheader(filenames[0])
        ra = header["RA"]
        dec = header["DEC"]

        for camname in ["east", "west"]:
            sources_cam = new_sources.loc[new_sources.camera == camname]

            cam_solution = astrometrynet_quick(
                {5200: "/data/astrometrynet/5200"},
                sources_cam,
                ra=ra,
                dec=dec,
                radius=5.0,
                pixel_scale=1.0,
                pixel_scale_factor_hi=1.2,
                pixel_scale_factor_lo=0.8,
                scales={5200: [5, 6]},
                plot=False,
                raise_on_unsolved=True,
                output_root=(output_root + f"_{camname}") if output_root else None,
                **kwargs,
            )

            new_hdu = fits.ImageHDU(name=camname.upper())
            if cam_solution.wcs:
                new_hdu.header += cam_solution.wcs.to_header(relax=True)
                new_hdu.header["SOLVED"] = True
            else:
                print(f"[WARNING]: Failed solving {camname}.")
                new_hdu.header["SOLVED"] = False
            hdul.append(new_hdu)

    hdul.writeto(str(proc_output_path), overwrite=True)

    return solution
